#include "GraphOutput.h"
#include <iostream>
#include <fstream>
const string GraphOutput::graph_file_suffix = ".graph";
const string GraphOutput::nodes_file_suffix = ".nodes";
const string GraphOutput::edges_file_suffix = ".edges";
const string GraphOutput::xml_file_suffix = ".xgmml";
const string GraphOutput::json_nodes_file_suffix = ".json_nodes";
const string GraphOutput::json_edges_file_suffix = ".json_edges";
const string GraphOutput::json_file_suffix = ".json";


void GraphOutput::init(bool erase){
  //printf("create a graph erase=%s graph_format=%d  first_id=%d\n", erase?"true":"false", graph_format, first_id);
  graph_file_name=(prefix+graph_file_suffix);
    nodes_file_name=(prefix+nodes_file_suffix);
    edges_file_name=(prefix+edges_file_suffix);
    xml_file_name=(prefix+xml_file_suffix);
    json_nodes_file_name=(prefix+json_nodes_file_suffix);
    json_edges_file_name=(prefix+json_edges_file_suffix);
    json_file_name=(prefix+json_file_suffix);

    if (graph_format==0) // FORMAT .GRAPH
    {
      graph_file = fopen(graph_file_name.c_str(),erase?"w":"a");
      fprintf(graph_file,"digraph debuijn {\n");
      return;
    }
    if (graph_format==1) // FORMAT .NODES AND .EDGES
    {
      nodes_file = fopen(nodes_file_name.c_str(),erase?"w":"a");
      edges_file = fopen(edges_file_name.c_str(),erase?"w":"a");
      return;
    } 
    if (graph_format==2) {// FORMAT .XGMML
      graph_file = fopen(xml_file_name.c_str(),erase?"w":"a"); 
      //fprintf(graph_file,"<?xml version=\"1.0\"?>\n<!DOCTYPE graph SYSTEM \"http://www.cs.rpi.edu/~puninj/XGMML/xgmml.dtd\">\n<graph directed=\"1\">\n");
      fprintf(graph_file,"<graph label=\"%s\"\n", (char *)prefix.c_str());
      fprintf(graph_file,"xmlns:dc=\"http://purl.org/dc/elements/1.1/\" \n");
      fprintf(graph_file,"xmlns:xlink=\"http://www.w3.org/1999/xlink\" \n");
      fprintf(graph_file,"xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\" \n");
      fprintf(graph_file,"xmlns:cy=\"http://www.cytoscape.org\" \n");
      fprintf(graph_file,"xmlns=\"http://www.cs.rpi.edu/XGMML\"  \n");
      fprintf(graph_file,"directed=\"1\">\n");
      return;
    }
    if (graph_format==3) {// FORMAT .json
      nodes_file = fopen(json_nodes_file_name.c_str(),erase?"w":"a");
      edges_file = fopen(json_edges_file_name.c_str(),erase?"w":"a");
      graph_file = fopen(json_file_name.c_str(),erase?"w":"a"); 
      return;
    }
}

// PIERRE: need something different than 0 for the first node
GraphOutput::GraphOutput(string prefix, int graph_format, long first_id)  : prefix(prefix), graph_format(graph_format), first_id(first_id)
{
  printf("graph_format=%d first_id=%d\n", graph_format, first_id);
  init(false);
}

GraphOutput::GraphOutput(string prefix, int graph_format) : prefix(prefix), graph_format(graph_format)
{
  first_id=0; 
  init(true);
}

void GraphOutput::close()
{
    if (graph_format==0)
    {
        fprintf(graph_file,"}\n");
        fclose(graph_file);
	return;
    }
    if (graph_format==1)
    {
        fclose(nodes_file);
        fclose(edges_file);
	return; 
    }
    if (graph_format==2)
      {
	fprintf(graph_file,"</graph>\n");
        fclose(graph_file);
	return; 
      }
    if (graph_format==3)
    {
      
      // We need to store all nodes and then all edges in the final .json file  
      fclose(nodes_file);
      fclose(edges_file);
      ifstream nodes(json_nodes_file_name.c_str(), ios::in);
      ifstream edges(json_edges_file_name.c_str(), ios::in);
      if(!edges || !nodes){fprintf(stderr,"Cannot open file %s or %s, exit\n", json_edges_file_suffix.c_str(), json_nodes_file_suffix.c_str()); exit(1);}
      string line;
      
      fprintf(graph_file,"{\n \"nodes\": [\n");      
      getline(nodes,line); 
      fprintf(graph_file,"%s",line.c_str()); // prints the first node without comma before
      while(getline(nodes,line)) fprintf(graph_file,",\n%s",line.c_str()); // prints the othe nodes
      fprintf(graph_file,"\n],\n");      
      nodes.close(); remove(json_nodes_file_name.c_str());
      
      fprintf(graph_file,"\"edges\": [\n");
      getline(edges,line); fprintf(graph_file,"%s",line.c_str()); // prints the first edge without comma before
      while(getline(edges,line)) fprintf(graph_file,",\n%s",line.c_str()); // prints the othe edges
      fprintf(graph_file,"\n]}\n");
      edges.close(); remove(json_edges_file_name.c_str());

      fclose(graph_file); 
      
      return; 
    }
}

// output a single node to a file
void GraphOutput::print_node(long index, char *ascii_node, char * comment)
{

  if (graph_format==0){ // DOT format
    fprintf(graph_file,"%ld [label=\"%s\"];\n",index,ascii_node);
    return;
  }
  if (graph_format==1){ // kissplice format
    fprintf(nodes_file,"%ld\t%s\n",index,ascii_node);
    return;
  }
  if (graph_format==2){ // XGMML format
    fprintf(graph_file,"<node id=\"%ld\" label=\"%s\">\n</node>\n",index,ascii_node);
    return;
  }
  if (graph_format==3){ // json format
    fprintf(nodes_file,"{ \"data\": { \"id\": %ld, \"sequence\": \"%s\" %s}}\n",index,ascii_node, comment);
    return;
  }
  
}

// output a single edge to a file
void GraphOutput::print_edge(long id, long id2, string label)
{
  if (graph_format==0){ // DOT format
    fprintf(graph_file,"%ld -> %ld [label=\"%s\"];\n",id,id2,label.c_str());
    return;
  }
  if (graph_format==1){ // kissplice format
    fprintf(edges_file,"%ld\t%ld\t%s\n",id,id2,label.c_str());
    return;
  }
  if (graph_format==2){ // XGMML format
    fprintf(graph_file,"<edge source=\"%ld\" target=\"%ld\" label=\"%s\">\n</edge>\n",id,id2,label.c_str());
    return;
  }
  if (graph_format==3){ // json format
    //fprintf(edges_file,"{ \"data\":{ \"id\": \"%ld%ld\", \"source\": \"%ld\",\"target\": \"%ld\",\"direction\": \"%s\"}}\n",id, id2, id,id2,label.c_str());
    fprintf(edges_file,"{ \"data\":{ \"source\": %ld,\"target\": %ld,\"direction\": \"%s\"}}\n",id,id2,label.c_str());
    return;
  }
  
}

void GraphOutput::load_nodes_extremities(string linear_seqs_name)
{
  kmer_links.clear(); // PIERRE: reset previous stored kmer links

    Bank *Nodes = new Bank((char *)linear_seqs_name.c_str());
    long nodeNb = first_id; //PIERRE;
    char * rseq;
    int readlen;

    sizeKmer--; // nodes extremities overlap by (k-1)-mers, so let's extract (k-1)-mers

    while (Nodes->get_next_seq(&rseq,&readlen))
    {
        kmer_type left_kmer, right_kmer, left_kmer_fw, left_kmer_rc, right_kmer_fw, right_kmer_rc;
        left_kmer = extractKmerFromRead(rseq,0,&left_kmer_fw,&left_kmer_rc, false);
        right_kmer = extractKmerFromRead(rseq,readlen-sizeKmer,&right_kmer_fw,&right_kmer_rc, false);
        Strand left_strand = (left_kmer == left_kmer_fw)?FW:RC;
        Strand right_strand = (right_kmer == right_kmer_fw)?FW:RC;

        kmer_links[left_kmer].insert(node_strand(nodeNb, left_strand, LEFT));
        kmer_links[right_kmer].insert(node_strand(nodeNb, right_strand, RIGHT));

        nodeNb++;
    }
    Nodes->close();
    delete Nodes;

    sizeKmer++; // make sure to restore k
}

int GraphOutput::construct_graph(string linear_seqs_name) // PIERRE: i need to know the last nodeNb
{
    Bank *Nodes = new Bank((char *)linear_seqs_name.c_str());
    long nodeNb = first_id; //PIERRE
    char * rseq;
    int readlen;
    long nb_edges = 0;

    
    while (Nodes->get_next_seq(&rseq,&readlen))
    {
      print_node(nodeNb, rseq, (char *)"");
      nodeNb++;
    }

    Nodes->rewind_all();
    nodeNb = first_id; // PIERRE

    sizeKmer--; // nodes extremities overlap by (k-1)-mers, so let's extract (k-1)-mers

    // for each node, output all the out-edges (in-edges will correspond to out-edges of neighbors)
    while (Nodes->get_next_seq(&rseq,&readlen))
    {

        kmer_type left_kmer, right_kmer, left_kmer_fw, left_kmer_rc, right_kmer_fw, right_kmer_rc;
        set<node_strand>::iterator it;

        left_kmer = extractKmerFromRead(rseq,0,&left_kmer_fw,&left_kmer_rc, false);
        right_kmer = extractKmerFromRead(rseq,readlen-sizeKmer,&right_kmer_fw,&right_kmer_rc, false);
        Strand left_strand = (left_kmer == left_kmer_fw)?FW:RC;
        Strand right_strand = (right_kmer == right_kmer_fw)?FW:RC;

        // left edges (are revcomp extensions)
        int nb_left_edges = 0;
        for (it = kmer_links[left_kmer].begin(); it != kmer_links[left_kmer].end(); it++)
        {
            long cur_node = it->node;
            Strand cur_strand = it->strand;
            LeftOrRight cur_left_or_right = it->left_or_right;

            if (cur_node == nodeNb) // prevent self loops on same kmer
                 if (readlen == sizeKmer)
                    continue;
            
            string label = "R";

            if (cur_left_or_right == LEFT)
            {
                if (cur_strand != left_strand)
                    label+=(string)"F";
                else
                    continue;
            }
            else
            {
                if (cur_strand == left_strand)
                    label+=(string)"R";
                else
                    continue;
            }

            print_edge(nodeNb,cur_node,label);
            nb_left_edges++;
        }


        // right edges
        int nb_right_edges = 0;
        for (it = kmer_links[right_kmer].begin(); it != kmer_links[right_kmer].end(); it++)
        {
            long cur_node = it->node;
            Strand cur_strand = it->strand;
            LeftOrRight cur_left_or_right = it->left_or_right;

            if (cur_node == nodeNb) // prevent self loops on same kmer
                 if (readlen == sizeKmer)
                    continue;
           
            string label = "F";

            if (cur_left_or_right == LEFT)
            {
                if (cur_strand == right_strand)
                    label+=(string)"F";
                else
                    continue;
            }
            else
            {
                if (cur_strand != right_strand)
                    label+=(string)"R";
                else
                    continue;
            }

            print_edge(nodeNb,cur_node,label);
            nb_right_edges++;
        }

        nb_edges += nb_left_edges + nb_right_edges;
            
        nodeNb++;
    }

    sizeKmer++; // make sure to restore k
    Nodes->close();
    delete Nodes;
    return nodeNb; // PIERRE
}


