#include <stdio.h>
#include <string.h>
#include <string>
#include <vector>
#include "LabelledCEdge.h"

#define MAX 1024

using namespace std;

struct NodeSeq {
  char *seq;
  int node;
  NodeSeq(char *initSeq, int initNode) : seq(initSeq), node(initNode) { }
  bool operator== (const NodeSeq& rhs) const { return (strcmp(seq, rhs.seq) == 0); }
  bool operator<  (const NodeSeq& rhs) const { return (strcmp(seq, rhs.seq) <  0); }
};

FILE* open_file( char* filename )
{
  FILE* file = fopen( filename, "r" );
  
  if ( file == NULL )
  {
    fprintf( stderr, "Problem opening %s!\n", filename );
    exit( EXIT_FAILURE );
  }
  
  return file;
}

int count_nb_lines( FILE* file )
{
  int ch, number_of_lines = 0;

  while (EOF != (ch=getc(file)))
    if ('\n' == ch)
      number_of_lines++;
  
  // Set the cursor back to the begining of the file.
  rewind(file);
  
  // Don't care if the last line has a '\n' or not. We over-estimate it.
  return number_of_lines + 1; 
}

void readNodeFile(char* nodes_fname, vector<NodeSeq>& nodes, int k_val )
{
  FILE* node_file = open_file(nodes_fname);
  char* buffer = new char[100 * MAX];
  char* seq;
  
  nodes.reserve(count_nb_lines(node_file));
  
  while ( fgets(buffer, 100 * MAX, node_file) != NULL )
  {
    char* p;

    if (strlen(buffer) == 100 * MAX)
    {  
      p = strtok(buffer, "\t\n");
      fprintf(stdout, "ERROR: node %s with sequence larger than %d!", p, 100 * MAX);
      exit(0);
    }
      
    // Node label, should be contiguous
    p = strtok( buffer, "\t\n" );
    int node = atoi(p);
          
    // Node seq
    p = strtok( NULL, "\t\n"  );

    // We only need the prefix of size k
    seq = new char[k_val + 1];
    strncpy( seq, p, k_val);

    nodes.push_back( NodeSeq(seq, node) );
  }
  
  delete [] buffer;
  fclose(node_file);
}

void readEdgeFile( char* nodes_fname, vector<LabelledCEdge>& edges )
{
  FILE *edge_file = open_file(nodes_fname);

  char* buffer = new char[100 * MAX];
  char* u = new char[MAX];
  char* v = new char[MAX];
  char* label = new char[MAX];
  
  edges.reserve(count_nb_lines(edge_file));

  while ( fgets(buffer, 100 * MAX, edge_file) != NULL )
  {
    char* p;

    // outgoing node
    p = strtok( buffer, "\t\n" );
    strcpy( u, p );

    // incoming node
    p = strtok( NULL, "\t\n" );
    strcpy( v, p );

    // edge label
    p = strtok( NULL, "\t\n" );
    strcpy(label, p);
  
    edges.push_back( LabelledCEdge( atoi(u), atoi(v), label ) );
  }
  
  sort( edges.begin(), edges.end() );
  
  delete [] buffer;
  delete [] u;
  delete [] v;
  delete [] label;
  
  fclose(edge_file);
}

static char complement(char b)
{
  switch(b)
    {
    case 'A': return 'T';
    case 'T': return 'A';
    case 'G': return 'C';
    case 'C': return 'G';
      
    case 'a': return 't';
    case 't': return 'a';
    case 'g': return 'c';
    case 'c': return 'g';
    
    case 'N': return 'N';
    case '*': return '*';  
    }
  return '?';
}

// rev_seq needs to be global! let's make sure we allocate only once
char *rev_seq = NULL;
char *revComp(char *seq)
{
  int size = strlen(seq);
  if (rev_seq == NULL)
    rev_seq = new char[size + 1];
   
  for (int i = size - 1; i >= 0; i--)
    rev_seq[(size - 1) - i] = complement(seq[i]); 
  
  return rev_seq;
}

int findNode(char *query, vector<NodeSeq>& nodes)
{
  vector<NodeSeq>::iterator low, low_r;
  char *query_r = revComp(query);

  // The value of "x" in NodeSeq(query, x) doesn't matter 
  low   = lower_bound(nodes.begin(), nodes.end(), NodeSeq(query, 0));
  low_r = lower_bound(nodes.begin(), nodes.end(), NodeSeq(query_r, 0));

  if (low != nodes.end() && *low == NodeSeq(query, 0))
    return low->node;
  if (low_r != nodes.end() && *low_r == NodeSeq(query_r, 0)) 
    return low_r->node;

  return -1;
}

void readCounts(char* counts_fname, vector<NodeSeq>& nodes, vector<int>& counts)
{
  FILE *count_file = open_file(counts_fname);
  char* buffer = new char[100 * MAX];
       
  while ( fgets(buffer, 100 * MAX, count_file) != NULL )
  {
    char* p;
     
    // Sequence
    p = strtok( buffer, "\t\n " );
    int node = findNode(p, nodes);

    // Count 
    p = strtok( NULL, "\t\n "  );
    if (node != -1)
      counts[node] += atoi(p);
  }
  
  delete [] buffer;
    
  fclose(count_file);
}

int findEdge(vector<LabelledCEdge>& allEdges, LabelledCEdge e)
{
  vector<LabelledCEdge>::iterator low; 

  low = lower_bound(allEdges.begin(), allEdges.end(), e);
  if (low != allEdges.end())
    return low - allEdges.begin();
  
  fprintf(stderr, "inconsistent graph!");
  return -1;
}

// Move this to CEdge
LabelledCEdge reverse(LabelledCEdge e)
{
  char *label = new char[3];
  
  label[0] = e.label[1] == 'F' ? 'R' : 'F';
  label[1] = e.label[0] == 'F' ? 'R' : 'F';
  label[2] = '\0';
  
  return LabelledCEdge(e.getSecond(), e.getFirst(), label);
}


void errorRemoval(vector<LabelledCEdge>& allEdges, int nbNodes, double cutoff, vector<bool>& removed, vector<int>& counts)
{
  int offset = 0;
  for ( int src = 0 ; src < nbNodes; src++ )
  {
    // Count the number of edges (they are ordered in all_edges) of node src
    int size = 0;
    while ( offset + size < (int) allEdges.size() && allEdges[offset + size].getFirst() == src )
      size++;
   
    // Compute the sum of coverage for each direction
    int sum_F = 0, sum_R = 0;
    for (int k = 0; k < size; k++)
      if (!removed[offset + k])
      {
	int target = allEdges[offset + k].getSecond();
	if (allEdges[offset + k].label[0] == 'F') 
	  sum_F += counts[target];
	
	if (allEdges[offset + k].label[0] == 'R')
	  sum_R += counts[target];
      }
    
    // Remove the edges with relative coverage below cutoff
    for (int k = 0; k < size; k++)
      if (!removed[offset + k])
      {
	int target = allEdges[offset + k].getSecond();
	double ratio;

	ratio = (double)counts[target] / (double)sum_F;
	if (allEdges[offset + k].label[0] == 'F' && ratio < cutoff)
	{  
	  removed[offset + k] = true;
	  removed[findEdge(allEdges, reverse(allEdges[offset+k]))] = true;
	}

	ratio = (double)counts[target] / (double)sum_R;
	if (allEdges[offset + k].label[0] == 'R' && ratio < cutoff)
	{
	  removed[offset + k] = true;
	  removed[findEdge(allEdges, reverse(allEdges[offset+k]))] = true;
	}
      }

    offset += size;
  }
}

int main( int argc, char** argv )
{
  string output_file = "./clean_graph.edges";

  if ( argc < 6 )
  { 
    fprintf( stderr, "Wrong number of arguments!\n" );
    fprintf( stderr, "Usage: ./error_removal edge_file node_file k_value k_mer_count_file cutoff [output_file]\n" );
    return 0;
  }
  
  if ( argc == 7 )
  {
    output_file = argv[6];
  }
  
  int k_value = atoi( argv[3] );
  double cutoff = atof( argv[5] );

  // Read edge file
  vector<LabelledCEdge> allEdges;
  readEdgeFile( argv[1], allEdges );

  // Read node file
  vector<NodeSeq> nodes;
  readNodeFile(argv[2], nodes, k_value );

  // Read count file
  vector<int> counts(nodes.size(), 0);
  sort(nodes.begin(), nodes.end());
  readCounts(argv[4], nodes, counts);

  vector<bool> removed(allEdges.size(), false);
  errorRemoval(allEdges, (int)nodes.size(), cutoff, removed, counts);
  
  int nb_removed = 0;
  FILE *output = fopen(output_file.c_str(), "w");
  for (int i = 0 ; i < (int)allEdges.size(); i++)
    if (!removed[i])
      fprintf(output, "%d\t%d\t%s\n", allEdges[i].getFirst(), allEdges[i].getSecond(), allEdges[i].label);
    else
      nb_removed++;
  fclose(output);
  
  fprintf(stdout, "%d out of %d edges removed\n", nb_removed, (int)allEdges.size());
 
  return 0;
}
