/* ***************************************************************************
 *
 *                              KisSplice
 *      de-novo calling alternative splicing events from RNA-seq data.
 *
 * ***************************************************************************
 *
 * Copyright INRIA
 *  contributors :  Vincent Lacroix
 *                  Pierre Peterlongo
 *                  Gustavo Sacomoto
 *                  Alice Julien-Laferriere
 *                  David Parsons
 *                  Vincent Miele
 *
 * pierre.peterlongo@inria.fr
 * vincent.lacroix@univ-lyon1.fr
 *
 * This software is a computer program whose purpose is to detect alternative
 * splicing events from RNA-seq data.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".

 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.

 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */



// ===========================================================================
//                               Include Libraries
// ===========================================================================
// ===========================================================================
//                             Include Project Files
// ===========================================================================
#include "TestNGraph.h"

// ===========================================================================
//                             Declare Used Namespaces
// ===========================================================================

//############################################################################
//                                                                           #
//                             Class Test_cgraph                             #
//                                                                           #
//############################################################################
CPPUNIT_TEST_SUITE_REGISTRATION( TestNGraph);

// ===========================================================================
//                               Static attributes
// ===========================================================================

// ===========================================================================
//                                  Constructors
// ===========================================================================
TestNGraph::TestNGraph(void)
{
}

// ===========================================================================
//                                  Destructors
// ===========================================================================

// ===========================================================================
//                                   Operators
// ===========================================================================

// ===========================================================================
//                                 Public Methods
// ===========================================================================
void TestNGraph::setUp(void) {
//===================== Graph 1 creation
  vector<int> labels;
  vector<LabelledCEdge> all_edges;
  int kValue = 5;
  //reading graph
  string base_name = "data/graphtest_k5";
  char edge_fname[1024];
  char nodes_fname[1024];
  sprintf( edge_fname, "%s.edges"  ,  base_name.c_str() );
  sprintf( nodes_fname,  "%s.nodes",  base_name.c_str());
  FILE* file = fopen(nodes_fname, "r"); // we suppose the opening is always a succes
  read_node_file( file, seqs, kValue);
  file = fopen(edge_fname, "r");
  read_edge_file( file, all_edges );

  // Map the node labels to contiguous ids
  CGraph graph_tmp( (int) seqs.size(), all_edges, kValue );
  //  bcc extraction:
  vector < vector<CEdge> > big_bcc = find_bcc(graph_tmp);
// Creation of the NGraph object
  graph1 = new NGraph(graph_tmp, seqs, all_edges, big_bcc[0]);

  //===================== Graph 2 creation
  vector<int> labels2;
  vector<LabelledCEdge> all_edges2;
  kValue = 3;
  base_name = "data/graphtest1_k3";
  sprintf( edge_fname, "%s.edges"  ,  base_name.c_str() );
  sprintf( nodes_fname,  "%s.nodes",  base_name.c_str());

  file = fopen(nodes_fname, "r");
  read_node_file( file, seqs2, kValue);
  file = fopen(edge_fname, "r");

  read_edge_file( file, all_edges2 );
  // Map the node labels to contiguous ids
  CGraph graph_tmp2( (int) seqs2.size(), all_edges2, kValue );
  //  bcc extraction:
  vector < vector<CEdge> > big_bcc2 = find_bcc(graph_tmp2);
  CPPUNIT_ASSERT(big_bcc2.size() == 1);
  // Creation of the NGraph object
  graph2 = new NGraph(graph_tmp2, seqs2, all_edges2, big_bcc2[0]);

  }

void TestNGraph::tearDown(void) {
    delete graph1;
    delete graph2;

}

void TestNGraph::all_compress(void) {
  graph1->compress_all_paths();
  // two lips compression
  CPPUNIT_ASSERT( graph1->getNodes().size() == 4 );

  FILE *log_file = tmpfile();

  int n_compressed_bubbles;
  graph1->compress_all_bubbles( &n_compressed_bubbles, log_file, 1 );
  // sequences of the two lips compressed into one
  CPPUNIT_ASSERT(graph1->getNodes().size() == 3);
//=============== graph2
  graph2->compress_all_paths();
  CPPUNIT_ASSERT( graph2->getNodes().size() == 8 );
  graph2->compress_all_bubbles( &n_compressed_bubbles, log_file, 2 );
  CPPUNIT_ASSERT( graph2->getNodes().size() == 7 );// compression of the snp path
  CPPUNIT_ASSERT( graph2->getNbOutput() == 1 );// compression of the snp path updating the counter
  fclose( log_file );
  // TODO: next step, in graph2 there is an as event to retrieve
}


// ===========================================================================
//                                Protected Methods
// ===========================================================================

// ===========================================================================
//                              Non inline accessors
// ===========================================================================
