/* ***************************************************************************
 *
 *                              KisSplice
 *      de-novo calling alternative splicing events from RNA-seq data.
 *
 * ***************************************************************************
 *
 * Copyright INRIA
 *  contributors :  Vincent Lacroix
 *                  Pierre Peterlongo
 *                  Gustavo Sacomoto
 *                  Vincent Miele
 *                  Alice Julien-Laferriere
 *                  David Parsons
 *
 * pierre.peterlongo@inria.fr
 * vincent.lacroix@univ-lyon1.fr
 *
 * This software is a computer program whose purpose is to detect alternative
 * splicing events from RNA-seq data.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * 'http://www.cecill.info'.

 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.

 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

// ===========================================================================
//                               Include Libraries
// ===========================================================================
// ===========================================================================
//                             Include Project Files
// ===========================================================================
#include "TestSplitBcc.h"

// ===========================================================================
//                             Declare Used Namespaces
// ===========================================================================

//############################################################################
//                                                                           #
//                             Class TestSplitBcc                          #
//                                                                           #
//############################################################################

CPPUNIT_TEST_SUITE_REGISTRATION( TestSplitBcc);

// ===========================================================================
//                               Static attributes
// ===========================================================================

// ===========================================================================
//                                  Constructors
// ===========================================================================
TestSplitBcc::TestSplitBcc(void) {
}

// ===========================================================================
//                                  Destructors
// ===========================================================================

// ===========================================================================
//                                   Operators
// ===========================================================================

// ===========================================================================
//                                 Public Methods
// ===========================================================================

//=================== Creating test environment
void TestSplitBcc::setUp(void) {
  /*  graphs were we want to find a bcc:
   *           0
   *     g1:  / \
   *         1---2
   *
   */
  char idLabel[]= "FR";
  vector<LabelledCEdge> labels;
  labels.push_back( LabelledCEdge( 0,  1 , idLabel));
  labels.push_back( LabelledCEdge( 0 , 2 , idLabel));
  labels.push_back( LabelledCEdge( 1 , 2 , idLabel));
  labels.push_back( LabelledCEdge( 2 , 0 , idLabel));
  labels.push_back( LabelledCEdge( 2 , 1 , idLabel));
  graph1 =  new CGraph(3, labels, 25);




  /*
   *       0
   * g2:  / \       3-4
   *     1---2
   *
   *two nodes bcc are supposed to be ignored in Gustavo's algorithm
   */
  labels.push_back( LabelledCEdge(3,4 , idLabel));
  labels.push_back( LabelledCEdge(4,3 , idLabel));
  graph2 = new CGraph(5, labels, 25);
  for ( int i = 0 ; i < (int)labels.size() ; i++ )
  {
    graph2->insert_unique_edge_dir( labels[i].getFirst(), labels[i].getSecond() );
  }



  /*
   *         0	    7
   *        /  \	/  \
   * g3:   1 -- 2     6
   *        \  /  \   |
   *         5	   3--4
   */
    vector<LabelledCEdge> labels2;
  labels2.push_back( LabelledCEdge( 0,  1 , idLabel));
  labels2.push_back( LabelledCEdge( 0 , 2 , idLabel));
  labels2.push_back( LabelledCEdge( 1 , 0 , idLabel));
  labels2.push_back( LabelledCEdge( 1 , 2 , idLabel));
  labels2.push_back( LabelledCEdge( 1 , 5 , idLabel));
  labels2.push_back( LabelledCEdge( 2 , 0 , idLabel));
  labels2.push_back( LabelledCEdge( 2 , 1 , idLabel));
  labels2.push_back( LabelledCEdge( 2 , 3 , idLabel));
  labels2.push_back( LabelledCEdge( 2 , 5 , idLabel));
  labels2.push_back( LabelledCEdge( 2 , 7 , idLabel));
  labels2.push_back( LabelledCEdge( 3 , 2 , idLabel));
  labels2.push_back( LabelledCEdge( 3 , 4 , idLabel));
  labels2.push_back( LabelledCEdge( 4 , 3 , idLabel));
  labels2.push_back( LabelledCEdge( 4 , 6 , idLabel));
  labels2.push_back( LabelledCEdge( 6 , 7 , idLabel));
  labels2.push_back( LabelledCEdge( 7 , 6 , idLabel));
  labels2.push_back( LabelledCEdge( 7 , 2 , idLabel));

  graph3 = new CGraph(8, labels2, 25);

}
//================== Destroying test environment
void TestSplitBcc::tearDown( void )
{
  delete graph1;
  delete graph2;
  delete graph3;
}

//================ Starting test =========================
void TestSplitBcc::testing_find_bcc( void )
{
  vector< vector<CEdge> > bcc;
  //========== testing graph1
  bcc = find_bcc(*graph1);
  CPPUNIT_ASSERT(bcc.size()==1);
  //========== testing  graph2
  bcc = find_bcc(*graph2);
  CPPUNIT_ASSERT(bcc.size()==1);
  //========== testing  graph3
  bcc = find_bcc(*graph3);
  CPPUNIT_ASSERT(bcc.size()==2);

}


// ===========================================================================
//                                Protected Methods
// ===========================================================================

// ===========================================================================
//                              Non inline accessors
// ===========================================================================

